@@ -28,112 +28,126 @@
 #include "cmds.h"
 #include "std_cmds.h"
 
-#ifdef SFTP_SUPPORT
-#include "secsftp_cmds.h"
-#endif
-
 extern int timeout;
-extern int serverport;
-extern struct loginstruct lsforward;
-char* portcmd;
+extern struct hostent_list* hostcache;
 
-struct loginfo li = { NULL, NULL, NULL, 0, 0 };
-struct conn_info_st conn_info = { 0, 0, 0, 0 };
+struct log_cmd_st lcs = { NULL, NULL, NULL, 0, 0 };
 struct cmdhandlerstruct *cmdhandler;
+struct conn_info_st conn_info = { 0, 0 };
 
 int checkforabort(struct clientinfo*);
 
-int handlecmds(const char* hostnames, struct clientinfo *clntinfo) {
+int handle_cmds(struct clientinfo *clntinfo) {
 	char *buffer = 0;
-	int ss, cs, ret;
+	int ss, cs;
 	int i;
 
-	conn_info.lsforward = &lsforward;
-	conn_info.li = &li;
+	conn_info.lcs = &lcs;
 	conn_info.clntinfo = clntinfo;
-
-	ss = clntinfo->serversocket = -1;
+	clntinfo->cachefd = -1;
+	log(9, "setting dataclientsock to -1 (initial)");
 	clntinfo->dataclientsock = clntinfo->dataserversock = -1;
-	conn_info.usergiven = 0;
-	portcmd = 0;
-	if ((ret = waitclient(hostnames, clntinfo)) < 0) {
-		/* An error occured */
-		return -1;
-	}
+	clntinfo->dataport = socketinfo_get_local_port(clntinfo->clientsocket) - 1;
+	log(9, "dataport is set to %d", clntinfo->dataport);
+	clntinfo->portcmd = (char*) 0;
+	ss = clntinfo->serversocket;
 	cs = clntinfo->clientsocket;
 
+	/*
+            Server host name                                     svrname
+            Server IP address                                    svrip
+            Server host name as specified in the login           svrlogin
+            Client host name                                     clntname
+            Client host IP address                               clntip
+            Interface IP to the client                           ifipclnt
+            Interface IP to the server                           ifipsvr
+	*/
 
-	li.user = 0;
-	li.name = (char*) 0;
+
+	lcs.clntip = strdup(get_char_ip(GET_IP_CLIENT, clntinfo));
+	enough_mem(lcs.clntip);
+	lcs.clntname = hostent_get_name(&hostcache, inet_addr(lcs.clntip));
+
+	lcs.ifipclnt = strdup(conv_ip_to_char(clntinfo->addr_to_client));
+	enough_mem(lcs.ifipclnt);
+
+	lcs.filename = (char*) 0;
 
 	while (1) {
 contin:
-#ifdef SFTP_SUPPORT
-		if (clntinfo->servermode == SECURE_SFTP) {
-			li.service = "sftp";
-			cmdhandler = &secsftp_cmdhandler[0];
-		} else {
-#endif
-			li.service = "ftp";
-			cmdhandler = &std_cmdhandler[0];
-#ifdef SFTP_SUPPORT
-		}
-#endif
+		lcs.service = "ftp";
+		cmdhandler = &std_cmdhandler[0];
+
 		ss = clntinfo->serversocket;
+
 		if (timeout) {
 			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
 			return -1;
 		}
 		if (buffer) {
 			/* log the command */
-			log_cmd(&li);
+			log_cmd(&lcs);
 			free(buffer);
+			free(lcs.method);
 			buffer = 0;
-			li.respcode = 0;
-			li.transferred = 0;
+			lcs.respcode = 0;
+			lcs.transferred = 0;
 		}
+
+		errno = 0;
 		buffer = readline(cs);
+
 		if (!buffer && timeout) {
 			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
 			err_time_readline(cs);
 			return -1;
 		}
 		if (!buffer) {
+			if (!errno) {
+				set_errstr("The client has probably closed the connection without saying QUIT");
+			} else {
+				set_errstr(strerror(errno));
+			}
 			err_readline(cs);
 			return -1;
 		}
-	
+
 		if (buffer[0] == '\0') {
 			/* empty line. Prevent logging of the command */
 			free(buffer);
 			buffer = 0;
 			goto contin;
 		}
-		
-		li.cmd = buffer;
+
+		lcs.cmd = buffer;
+
+		i = 0;
+		lcs.method = quotstrtok(lcs.cmd, WHITESPACES, &i);
+
+		/* check for passcmds and dontpasscmds options */
+		if (passcmd_check(lcs.method) == 0) {
+			say(clntinfo->clientsocket,
+				"500 Command not implemented\r\n");
+			log(8, "rejected command %s", lcs.method);
+			goto contin;
+		}
 
 		/* Retrieve a file */
 		clntinfo->mode = RETR;
 		i = 0;
 
-		/* If we are not yet connected, the only valid commands are
-		 * USER and PASS */
-		if (ss == -1 && !checkbegin(buffer, "USER ")
-			     && !checkbegin(buffer, "PASS ")
-			     && !checkbegin(buffer, "QUIT")) {
-			say(cs, "530 Please login with USER and PASS.\r\n");
-			li.respcode = 530;
-			goto contin;
-		}
-
 		while (cmdhandler[i].cmd) {
 			if (checkbegin(buffer, cmdhandler[i].cmd)) {
 				int ret = (cmdhandler[i].func)
 						(buffer, &conn_info);
+				if (checkbegin(buffer, "PASS ")) {
+					memset(buffer, 0, strlen(buffer));
+				}
 				switch (ret) {
 					case CMD_PASS:
 						if (passcmd(buffer, clntinfo)
 								< 0) {
+							free(lcs.method);
 							free(buffer);
 							return -1;
 						}
@@ -141,8 +155,21 @@ contin:
 					case CMD_HANDLED:
 						break;
 					case CMD_QUIT:
+						transfer_cleanup(
+							conn_info.clntinfo);
+						free(lcs.method);
 						free(buffer);
 						return 0;
+					case CMD_ABORT:
+						transfer_cleanup(
+							conn_info.clntinfo);
+						free(lcs.method);
+						free(buffer);
+						return -1;
+					case CMD_ERROR:
+						transfer_cleanup(
+							conn_info.clntinfo);
+						break;
 				}
 				/* found and called proper function */
 				goto contin;
@@ -154,361 +181,156 @@ contin:
 
 		if (cmdhandler == &std_cmdhandler[0]) {
 			if (passcmd(buffer, clntinfo) < 0) {
+				free(lcs.method);
 				free(buffer);
 				return -1;
 			}
 		} else {
 			say(clntinfo->clientsocket, 
-			"500 Command not understood or not implemented.\r\n");
+					"500 Command not implemented.\r\n");
 		}
+		/* end of while loop */
 	}
-	/* li.client, li.server and li.user  are freed at the termination of
-	 * the programm */
-}
-
-
-
-/* getuserdest() parses the "USER xyz" login string */
-
-#define IPBUFSIZE 16
-int getuserdest(char* buffer, struct clientinfo *clntinfo) {
-	char* dest, *user, *end, *orig_login;
-	char* buffnew = 0;
-	const char* port, *opt;
-	int i, f, cs;
-	i = 5;
-	f = 0;
-	cs = clntinfo->clientsocket;
-	clntinfo->servermode = getservermode();
-#ifdef NCFTP_TYPE5
-	log(9, "squeezing @");
-	char_squeeze(buffer, '@');
-#endif
-	while (buffer[i] == ' ')
-		i++;
-	if(i > strlen(buffer)) {
-		say(cs, "500 'USER': command requires a parameter.\r\n");
-		li.respcode = 500;
-		return 1;
-	}
-
-	user = clntinfo->user = (char*) malloc(strlen(buffer) + 1);
-	enough_mem(user);
-	dest = clntinfo->destination = (char*) malloc(strlen(buffer)+1);
-	enough_mem(dest);
-	clntinfo->destinationport = 0;
-
-	opt = get_option("passallauth");
-	if (opt) {
-		size_t bufsize;
-		/* passallauth specified */
-		/* transform buffer into "buffer"@targethost,21 */
-
-		orig_login = strdup(&buffer[i]);
-		enough_mem(orig_login);
-		bufsize =  strlen("USER ")
-				+ 1                         /* " */
-				+ strlen(orig_login)        /* orig_login */
-				+ 1                         /* " */
-				+ 1                         /* @ */
-				+ strlen(opt)               /* opt */
-				+ 1;                        /* Termination */
-		buffer = (char*) realloc(buffer, bufsize);
-		snprintf(buffer, bufsize, "USER \"%s\"@%s", orig_login, opt);
-		free(orig_login);
-		orig_login = 0;
-	}
-	
-	if (get_forward(&buffer[i], &lsforward) == 0) {
-		/* fill at least fwlogin in the structure */
-		lsforward.fwlogin = extract_userhost(&buffer[i]);
-	} else {
-		/* parse username from buffer[i] */
-		lsforward.fwuser = extract_username(&buffer[i]);
-		buffer = lsforward.login;
-		user = clntinfo->user = 
-			(char*) realloc(user, strlen(buffer) + 1);
-		enough_mem(user);
-		dest = clntinfo->destination =
-			(char*) realloc(dest, strlen(buffer) + 1);
-		enough_mem(dest);
-
-		i = 0;
-	}
-
-	if (buffer[i] == '\"') {
-		/* everything between two "" is a username
-		 * useful when logging in to a proxy */
-		end = strrchr(&buffer[i+1], '\"');
-		
-		if (end) {
-			/* quotation marks balanced */
-		
-			/* f specifies the length of the username */
-			f = end - &buffer[i+1];
-			strncpy(user, &buffer[i+1], f);
-			i += f + 2 /* quotation marks */;
-
-			/* The next character after the closing
-			 * quotation mark should be a @ or , so the
-			 * while loop is not executed and user gets
-			 * null-terminated. If it is not, the
-			 * characters are appended to user[] */
-		}
-	}
-	while(buffer[i] != '@' && buffer[i] != ',' 
-			&& i < strlen(buffer)) {
-		user[f] = buffer[i];
-		i++; f++;
-	}
-	user[f] = '\0';
-	i++;  /* shift over @ or , */
-	if (buffer[i-1] == '\0') {
-		const char* tmp;
-		const char* addtarget = 0;
-		int transproxy = 1;
-		char ipbuf[ IPBUFSIZE ];
-		/* first, lets terminate user. Since it is the last
-		 * value in the line, it is followed by an \r\n */
-		while(user[--f] < 32 && f >= 0) {
-			user[f] = '\0';
-		}
-		tmp = get_option("defaultforward");
-		if (!tmp) {
-			if (get_option("transparent-proxy")) {
-				if (strcmp(get_option("transparent-proxy"),
-							"on") == 0
-					||
-				    strcmp(get_option("transparent-proxy"),
-							"1") == 0
-					||
-				    strcmp(get_option("transparent-proxy"),
-							"yes") == 0) {
-					transproxy = 1;
-				} else {
-					transproxy = 0;
-				}
-			} else {
-				transproxy = 0;
-			}
-		}
-		if (transproxy) {
-			struct sockaddr_in sin;
-			const char* fwhost = 0;
-#ifdef HAVE_SOCKLEN_T
-		        socklen_t namelen;
-#else
- 			int namelen;
-#endif
-        		namelen = sizeof(sin);
-
-			log(6, "No defaultroute specified, acting as transparent proxy");
-			sin.sin_addr.s_addr = get_showaddr(clntinfo->clientsocket, SHOW_ORIGINAL_IP);
-			strncpy(ipbuf, inet_ntoa(*(struct in_addr *)
-						&sin.sin_addr), IPBUFSIZE);
-			tmp = ipbuf;
-			
-			/* Determine the IP the client sees of us */
-			if (getpeername(clntinfo->clientsocket,
-				(struct sockaddr *) &sin, &namelen) != 0) {
-				log(3, "Could not get peername for TP");
-			}
-			clntinfo->clientaddr = get_our_addr_to(
-					sin.sin_addr.s_addr, sin.sin_port);
-			sin.sin_addr.s_addr = clntinfo->clientaddr;
-			log(8, "Our IP to the client is %s", inet_ntoa(sin.sin_addr));
-
-			if ((fwhost = get_option("transparent-forward"))) {
-				/* user => user@targethost */
-				/* tmp => dest => forwardhost */
-
-				user = realloc(user, strlen(user) 
-						+ 1 + strlen(ipbuf) + 1);
-				enough_mem(user);
-				clntinfo->user = user;
-				strcat(user, "@");
-				strcat(user, ipbuf);
-
-				addtarget = fwhost;
-				log(8, "Using transparent proxy. Connecting to %s. User: %s", addtarget, user);
-			} else {
-				size_t destsize;
-				log(8, "Connecting to %s as transparent proxy",
-					tmp);
-				/* resize dest */
-				destsize = strlen(tmp) + 1;
-				dest = (char*) realloc(dest, destsize);
-				enough_mem(dest);
-				clntinfo->destination = dest;
-				snprintf(dest, destsize, "%s", tmp);
-				/* we don't have to free tmp */
-				addtarget = 0;
-			}
-		} else {
-			if (!tmp) {
-				/* No defaultroute specified and no
-				 * transparent proxy support. Kick it */
-				say(cs, "500 You are not allowed to connect\r\n");
-				return -1;
-			}
-			addtarget = tmp;
-		}
-		log(9, "Addtarget: %s", addtarget);
-		if (addtarget) {
-			/* copy the option of defaultforward into buffer */
-			size_t buffnewsize = strlen(buffer)
-						+ strlen(addtarget)
-						+ 1
-						+ 1;
-			buffnew = (char*) malloc(buffnewsize);
-			enough_mem(buffnew);
-			dest = (char*) realloc(dest, strlen(addtarget) + 1);
-			clntinfo->destination = dest;
-			snprintf(buffnew, buffnewsize, "%s@%s",
-					buffer, addtarget);
-			buffer = buffnew;
-			/* we don't have to free addtarget */
-			log(9, "Appended defaultforward option. New login string: %s. buffer[i]: %s", buffer, buffer+i);
-		}
-	}
-	if (buffer[i-1] != '\0') {
-		f =0;
-		while(buffer[i] && i < strlen(buffer)
-				&& buffer[i] != ','
-				&& buffer[i] != ':') {
-			dest[f] = buffer[i];
-			i++; f++;
-		}
-		dest[f] = '\0';
-		while (f > 0 && dest[--f] < 32) {
-			dest[f] = '\0';
-		}
-		port = strchr(buffer + i, ',');
-		if (!port) {
-			port = strchr(buffer + i, ':');
-		}
-		if (port) {
-			/* There is either a port specification
-			 * (decimal number) or server/client mode
-			 * specification */
-		
-			/* shift over the : or , */
-			port++;
-			if (0 == strchr("aps", *port))
-				/* *port != 'a' && *port != 'p') */ {
-				log(9, "analysing port: %s", port);
-				clntinfo->destinationport = atoi(port);
-				port = strchr(buffer + i + 1, ',');
-				if (!port) {
-					port = strchr(buffer + i + 1,
-							':');
-				}
-				if (port) {
-					port++;
-				}
-			}
-#ifdef SFTP_SUPPORT
-			if (port && (strchr("aps", *port))) {
-#else
-			if (port && (strchr("ap", *port))) {
-#endif
-				log(9, "mode specified: %s", port);
-				switch (*port) {
-					case 'a': 
-					clntinfo->servermode = ACTIVE;
-					log(9, "p-s: active");
-					break;
-					case 'p':
-					clntinfo->servermode = PASSIVE;
-					log(9, "p-s: passive");
-					break;
-#ifdef SFTP_SUPPORT
-					case 's':
-					clntinfo->servermode = SECURE_SFTP;
-					log(9, "p-s: secure-sftp");
-					break;
-#endif
-				}
-			}
-		}
-	}
-	if (buffnew) {
-		free(buffnew);
-		buffnew = 0;
-	}
-	if (!lsforward.fwuser) {
-		/* lsforward.fwuser = (char*) malloc(strlen(user) + 1);
-		strcpy(lsforward.fwuser, user); */
-		lsforward.fwuser = extract_username(user);
-	}
-	li.user = strdup(lsforward.fwuser);
-	enough_mem(li.user);
-	if (!clntinfo->destinationport) {
-		clntinfo->destinationport = getserverport();
-	}
-	log(7, "Client logged in: User: %s, Dest: %s", user, dest);
-	switch (clntinfo->servermode) {
-		case SECURE_SFTP:
-			log(8, "Secure sftp mode");
-			break;
-		default:
-			log(8, "Standard FTP mode, port: %d",
-				clntinfo->destinationport);
-	}
-
-	/* Get the Password */
+	/* lcs.host and lcs.user  are freed at the termination of the programm */
 	return 0;
 }
 
 
-int getpasswd(char* buffer, struct clientinfo *clntinfo) {
-	int i, f, cs;
-	char* pass;
-	pass = clntinfo->pass = (char*) malloc(strlen(buffer) + 1);
-	cs = clntinfo->clientsocket;
-	
-	f = strlen("PASS ");
-	i = 0;
-	while (buffer[f] == ' ') {
-		f++;
+char* getftpwd(struct clientinfo *clntinfo) {
+	char* answer, *dir;
+	char* dirstart, *dirend;
+	size_t dirsize;
+
+	say(clntinfo->serversocket, "PWD\r\n");
+	answer = ftp_readline(clntinfo->serversocket);
+	/*  257 "/home/joe" is current directory.  */
+	/*  257 "/home/joe/hi""ho lo di'da" is current directory. */
+
+	if ( ! checkdigits(answer, 257)) {
+		log(4, "PWD failed: %s", answer);
+		free(answer);
+		return (char*) 0;
 	}
-#ifdef NCFTP_TYPE5
-	log(9, "chopping first @ in password: %s", &buffer[f]);
-	while (buffer[f] == '@') {
-		f++;
+	dirstart = strchr(answer, '"');
+	dirend = strrchr(answer, '"');
+
+	if ( ! dirstart  ||  ! dirend ) {
+		log(4, "Could not parse PWD command: %s", answer);
+		free(answer);
+		return (char*) 0;
 	}
-	log(9, "after chopping: %s", &buffer[f]);
-#endif
-	while (buffer[f]) {
-		pass[i] = buffer[f];
-		i++; f++;
-	}
-	memset(buffer + f, 0, strlen(buffer) - f);
-	pass[i] = '\0';
-	while (i > 0 && pass[i] < 32) {
-		pass[i] = '\0';
-		i--;
-	}
-	return 0;
+
+	dirstart++;
+	dirend--;
+
+	dirsize = dirend - dirstart + 1 + 1;
+
+	dir = (char*) malloc(dirsize);
+	enough_mem(dir);
+
+	snprintf(dir, dirsize, "%s", dirstart);
+	free(answer);
+	char_squeeze(dir, '"');
+
+	return dir;
 }
 
+time_t getftpmdtm(const char* filename, struct clientinfo *clntinfo) {
+/*
+ *	ftp> quote mdtm bla
+ *	213 20010224102705
+ */
+	size_t cmdsize = strlen("MDTM ") + strlen(filename) + 2 + 1;
+	struct tm tms;
+	int i;
+	char* answer;
+	char* cmd = (char*) malloc(cmdsize);
 
-int passcmd(char* buffer, struct clientinfo *clntinfo) {
+	enough_mem(cmd);
+	snprintf(cmd, cmdsize, "MDTM %s\r\n", filename);
+
+	memset(&tms, 0, sizeof(tms));
+
+	say(clntinfo->serversocket, cmd);
+	free(cmd);
+	answer = ftp_readline(clntinfo->serversocket);
+	if ( ! checkdigits(answer, 213)) {
+		log(4, "Error reading MDTM answer: %s", answer);
+		free(answer);
+		return (time_t)(-1);
+	}
+
+	/* 213 20010224102705 */
+	i = sscanf(answer, "213 %4d%2d%2d%2d%2d%2d",
+			&tms.tm_year, &tms.tm_mon, &tms.tm_mday,
+			&tms.tm_hour, &tms.tm_min, &tms.tm_sec);
+
+	if (i != 6) {
+		log(4, "Error parsing MDTM answer: %s", answer);
+		free(answer);
+		return (time_t)(-1);
+	}
+	tms.tm_year -= 1900;  /* tm_year contains the number of years */
+			      /* since 1900 */
+	tms.tm_mon -= 1;      /* tm_mon starts with 0 */
+
+	free(answer);
+	return mktime( &tms );
+}
+
+unsigned long int getftpsize(char* filename, struct clientinfo *clntinfo) {
+/*
+ * ftp> quote size speak.ps
+ * 213 146617
+ */
+	size_t cmdsize = strlen("SIZE ") + strlen(filename) + 2 + 1;
+	unsigned long int size;
+	int i;
+	char* answer;
+	char* cmd = (char*) malloc(cmdsize);
+
+	enough_mem(cmd);
+	snprintf(cmd, cmdsize, "SIZE %s\r\n", filename);
+
+	say(clntinfo->serversocket, cmd);
+	free(cmd);
+	answer = ftp_readline(clntinfo->serversocket);
+	if ( ! checkdigits(answer, 213)) {
+		log(4, "Error reading SIZE answer: %s", answer);
+		free(answer);
+		return 0;
+	}
+
+	i = sscanf(answer, "213 %ld", &size);
+	if (i != 1) {
+		log(4, "Error parsing SIZE answer: %s", answer);
+		free(answer);
+		return 0;
+	}
+
+	free(answer);
+	return size;
+}
+
+int passcmd(const char* buffer, struct clientinfo *clntinfo) {
 	int cs = clntinfo->clientsocket;
 	int ss = clntinfo->serversocket;
 	char* sendbuf =0;
 	char *last = 0;
-	int ret;
 	size_t sendbufsize;
-	int transfer_start;
 
 	sendbufsize = strlen(buffer) + 3;
 	sendbuf = (char*) malloc(sendbufsize);
 	snprintf(sendbuf, sendbufsize, "%s\r\n", buffer);
+	log(9, "Send (server - %d): %s", ss, sendbuf);
 	say(ss, sendbuf);
 	free(sendbuf);
-	li.complete = 0;
+	lcs.complete = 0;
 	last = passall(ss, cs);
+	if (last) {
+		lcs.respcode = getcode(last);
+		log(9, "Send (client - %d): %s", cs, last);
+	}
 	if (!last) {
 		if (timeout) {
 			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
@@ -518,53 +340,35 @@ int passcmd(char* buffer, struct clientinfo *clntinfo) {
 		}
 		return -1;
 	}
-	if (checkdigits(last, 150) || checkdigits(last, 125)) {
-
-		if (portcmd) {
-			ret = activeclient(portcmd, clntinfo);
-			free(portcmd);
-			portcmd = 0;
-			if (ret) {
-				log(2, "Error in activeclient()");
-			}
-		}
-
-		if (initiate_transfer(clntinfo) < 0) {
-			free(last);
-			return -1;
-		}
-
-		transfer_start = time(NULL);
-		ret = transmitfile(clntinfo);
-		li.transfer_duration = time(NULL) - transfer_start;
-
-		if (!ret) {
-			char* t = passall(ss, cs);
-			li.respcode = respcode(t);
-			free(t);
-		}
-		else {
-			if (ret == 1) { /* generate error message */
-				if (timeout) {
-			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
-					err_time_readline(cs);
-				} else {
-					err_readline(cs);
-				}
-			}
-			li.respcode = respcode(last);
-			free(last);
-			return -1;
-		}
-	}
 	free(/*at*/ last);  /*  ;-)  */
-	last = 0;
-	li.complete = li.respcode == 226;
+	last                     =  (char*) 0;
+	lcs.complete = lcs.respcode == 226;
+	return 0;
+}
+
+int transfer_cleanup(struct clientinfo *clntinfo) {
+	if (clntinfo->portcmd) {
+		free(clntinfo->portcmd);
+		clntinfo->portcmd = (char*) 0;
+	}
+	if (clntinfo->dataserversock != -1) {
+		close(clntinfo->dataserversock);
+	}
+	if (clntinfo->dataclientsock != -1) {
+		close(clntinfo->dataclientsock);
+	}
 	return 0;
 }
 
 
-int initiate_transfer(struct clientinfo *clntinfo) {
+/* transfer_negotiate:
+ *
+ * return values:
+ *
+ * 	-1    dramatic error that should cause the program to terminate
+ * 	-2    non dramatic error
+ */
+int transfer_negotiate(struct clientinfo *clntinfo) {
 	int cs = clntinfo->clientsocket;
 	int ret;
 #ifdef HAVE_SOCKLEN_T
@@ -573,32 +377,62 @@ int initiate_transfer(struct clientinfo *clntinfo) {
 	int count;
 #endif
 	fd_set acc_set;
-	struct timeval tmo = { 300, 0 };
+	int servacc = 0, clntacc = 0;
+	struct timeval tmo;
 	struct sockaddr_in sin;
 
 	count = sizeof(sin);
-	
+	tmo.tv_sec = config_get_ioption("transfertimeout", 300);
+	tmo.tv_usec = 0;
+
+	if (clntinfo->portcmd) {
+		ret = activeclient(clntinfo->portcmd, clntinfo);
+		free(clntinfo->portcmd);
+		clntinfo->portcmd = (char*) 0;
+		if (ret) {
+			log(2, "Error in activeclient()");
+		}
+	}
+
+	if (clntinfo->mode == RETR && clntinfo->fromcache == 1) {
+		/* the data comes from the cache */
+		close(clntinfo->dataserversock);
+		clntinfo->dataserversock = clntinfo->cachefd;
+		servacc = 1;
+	}
+
+	/* if we're talking to the server in active ftp mode, the server
+	 * connects to us. Since we're talking to the client in passive
+	 * mode, the client connects to us, too! */
 	if (clntinfo->servermode == ACTIVE 
 	    && clntinfo->clientmode == PASSIVE) {
 		/* accept connections from both sides */
-		int servacc = 0, clntacc = 0;
+		clntinfo->waitforconnect = (int*) 0;
+		/* loop until both have connected */
 		while (servacc == 0 || clntacc == 0) {
 			FD_ZERO(&acc_set);
-			FD_SET(clntinfo->dataserversock, &acc_set);
-			FD_SET(clntinfo->dataclientsock, &acc_set);
+			if ( ! servacc ) {
+				FD_SET(clntinfo->dataserversock, &acc_set);
+			}
+			if ( ! clntacc ) {
+				FD_SET(clntinfo->dataclientsock, &acc_set);
+			}
+			/* select() and wait for both */
 			ret = select(MAX_VAL(clntinfo->dataserversock,
 					 clntinfo->dataclientsock) + 1,
 					&acc_set, NULL, NULL, &tmo);
 			if (ret < 0) {
-				log(2, "Select() error");
+				log(2, "Select() error1: %s", strerror(errno));
 				return -1;
 			}
 			if (ret == 0) {
 				log(2, "Connection timed out");
-				li.respcode = 500;
+				lcs.respcode = 500;
 				say(cs, "500 Connection timed out\r\n");
 				return -1;
 			}
+			/* no timeout, no error, one of them must have
+			 * connected, check if it was the server. */
 			if (FD_ISSET(clntinfo->dataserversock, &acc_set)) {
 				ret = accept(clntinfo->dataserversock,
 					(struct sockaddr*) &sin, &count);
@@ -611,6 +445,7 @@ int initiate_transfer(struct clientinfo *clntinfo) {
 				clntinfo->dataserversock = ret;
 				servacc = 1;
 			}
+			/* check if the client connected */
 			if (FD_ISSET(clntinfo->dataclientsock, &acc_set)) {
 				ret = accept(clntinfo->dataclientsock,
 					(struct sockaddr*) &sin, &count);
@@ -621,26 +456,37 @@ int initiate_transfer(struct clientinfo *clntinfo) {
 				}
 				close(clntinfo->dataclientsock);
 				clntinfo->dataclientsock = ret;
-				clntinfo->waitforconnect = 0;
 				clntacc = 1;
 			}
 		}
 	}
+	if (clntinfo->mode == RETR
+		    && clntinfo->fromcache == 1
+		    && clntinfo->waitforconnect == &clntinfo->dataserversock) {
+		/* the data comes from the cache */
+		clntinfo->waitforconnect = (int*) 0;
+/*		close(clntinfo->dataserversock);
+		clntinfo->dataserversock = clntinfo->cachefd;
+*/
+	}
 	if (clntinfo->waitforconnect) {
+		FD_ZERO(&acc_set);
+		if (*clntinfo->waitforconnect < 0) {
+			return -2;
+		}
 		FD_SET(*clntinfo->waitforconnect, &acc_set);
 		ret = select(*clntinfo->waitforconnect + 1, &acc_set,
 				NULL, NULL, &tmo);
 		if (ret < 0) {
-			log(2, "Select() error");
+			log(2, "Select() error2: %s", strerror(errno));
 			return -1;
 		}
 		if (ret == 0) {
 			log(2, "Connection timed out");
-			li.respcode = 500;
+			lcs.respcode = 500;
 			say(cs, "500 Connection timed out\r\n");
 			return -1;
 		}
-
 		/* accept connection from one side */
 		ret = accept(*clntinfo->waitforconnect,
 				(struct sockaddr*) &sin, &count);
@@ -652,7 +498,7 @@ int initiate_transfer(struct clientinfo *clntinfo) {
 		close(*clntinfo->waitforconnect);
 		*clntinfo->waitforconnect = ret;
 	}
-	
+
 	if (clntinfo->mode == STOR) {
 		/* swap descriptors */
 		ret = clntinfo->dataserversock;
@@ -667,28 +513,40 @@ int initiate_transfer(struct clientinfo *clntinfo) {
 #else
 #define TRANSMITBUFSIZE  PIPE_BUF
 #endif
-int transmitfile(struct clientinfo *clntinfo)  {
+int transfer_transmit(struct clientinfo *clntinfo)  {
 	char* buffer = (char*) malloc(TRANSMITBUFSIZE);
 	char* pbuf = 0;
-	int count = 0, nwritten, totwritten, sret = 0, scret = 0;
+	int count = 0;
+	int nwritten = 0, cachewritten, totwritten, sret = 0, scret = 0;
+	int cachefail = 0;
 	int cs = clntinfo->clientsocket;
-	/* error == 0 : success
-	 * error == 1 : fail, generate error message
-	 * error == 2 : fail, error message already generated
-	 */
-	int n, ret, error = 0;
+	int n, ret, error = 0, aborted = 0;
+	int maxfd;
 	int fdflags;
 	fd_set readset, writeset, exceptset;
-	struct timeval readtime = { 300, 0 };
-	struct timeval writetime = { 300, 0 }; 
+	struct timeval readtime;
+	struct timeval writetime;
 	time_t start, done, delay;
 	sigset_t sigset, oldset;
 
+	readtime.tv_sec = config_get_ioption("transfertimeout", 300);
+	readtime.tv_usec = 0;
+	writetime.tv_sec = config_get_ioption("transfertimeout", 300);
+	writetime.tv_usec = 0;
+
 	enough_mem(buffer);
 	FD_ZERO(&readset);
 	FD_ZERO(&writeset);
 	FD_ZERO(&exceptset);
-	FD_SET(clntinfo->dataserversock, &readset);
+
+	if (clntinfo->dataclientsock < 0) {
+		free(buffer);
+		return 0;
+	}
+
+	if (clntinfo->fromcache == 0) {
+		FD_SET(clntinfo->dataserversock, &readset);
+	}
 	FD_SET(cs, &exceptset);
 	FD_SET(cs, &readset);
 	FD_SET(clntinfo->dataclientsock, &writeset);
@@ -716,12 +574,24 @@ int transmitfile(struct clientinfo *clntinfo)  {
 		log(3, "sigprocmask() error: %s", strerror(errno));
 	}
 
-	/* read from the clientsocket as well as from the dataserversocket */
-	/* no writeset is specified here. The client could send an abort (it
-	 * is also in exceptset) and the server sends data */
-	while((sret = select(MAX_VAL(clntinfo->dataserversock, cs) + 1,
-				&readset, NULL, &exceptset,
-				&readtime) > 0)) {
+	while(1) {
+		/* Linux modifies the timeout struct */
+		readtime.tv_sec = config_get_ioption("transfertimeout", 300);
+		readtime.tv_usec = 0;
+		writetime.tv_sec = config_get_ioption("transfertimeout", 300);
+		writetime.tv_usec = 0;
+		count = -1; /* choose a number != 0 for the check below the
+			       while loop */
+
+		if (clntinfo->fromcache) {
+			readtime.tv_sec = 0;
+			readtime.tv_usec = 0;
+			maxfd = cs;
+		} else {
+			maxfd = MAX_VAL(clntinfo->dataserversock, cs);
+		}
+		sret = select(maxfd+1, &readset, NULL, &exceptset, &readtime);
+
 		/* save the errno value of select from sigprocmask() */
 		n = errno;
 		ret = sigprocmask(SIG_UNBLOCK, &sigset, &oldset);
@@ -731,8 +601,14 @@ int transmitfile(struct clientinfo *clntinfo)  {
 		}
 		errno = n;
 
+		if (sret <= 0) {
+			break;
+		}
+
 		/* Can we read data from the client ? */
-		if (FD_ISSET(clntinfo->dataserversock, &readset)) {
+		if (clntinfo->fromcache
+			|| FD_ISSET(clntinfo->dataserversock, &readset)) {
+
 			count = read(clntinfo->dataserversock,
 				buffer, TRANSMITBUFSIZE);
 			if (count == 0) {
@@ -740,32 +616,14 @@ int transmitfile(struct clientinfo *clntinfo)  {
 				break;
 			}
 			if (count < 0) {
-				log(3, "read error: %s", strerror(errno));
+				int err = errno;
+				log(3, "read error: %s", strerror(err));
+				set_errstr(strerror(err));
 				/* comm */
 				sret = 1;
-				error = 1;
+				error = TRNSMT_ERROR;
 				break;
 			}
-			/* This is only needed for sftp mode connections. If
-			 * the proxy acts between a real FTP client and a
-			 * real FTP server, both will do the job of
-			 * converting */
-			if (clntinfo->havetoconvert != CONV_NOTCONVERT) {
-				char* tmp;
-				if (clntinfo->havetoconvert == CONV_TOASCII) {
-					tmp = to_ascii(buffer, &count);
-					if (count > TRANSMITBUFSIZE) {
-						buffer = realloc(buffer, count);
-					}
-					memcpy((void*) buffer,
-						(void*) tmp, count);
-					free(tmp);
-				}
-				if (clntinfo->havetoconvert == CONV_FRMASCII) {
-					tmp = from_ascii(buffer, &count);
-					buffer = tmp;
-				}
-			}
 
 			/* comm */
 			pbuf = buffer;
@@ -802,10 +660,42 @@ int transmitfile(struct clientinfo *clntinfo)  {
 					break;
 				}
 				/* otherwise the descriptor must be ready */
+				/* write to the cache first */
+				if ( clntinfo->tocache &&  ! cachefail ) {
+					cachewritten = write(clntinfo->cachefd,
+						pbuf, count);
+					if (cachewritten != count) {
+						log(3, "Error writing to the "
+							"cache: %s",
+							strerror(errno));
+						cachefail = 1;
+					}
+				}
+				/* convert if we have to */
+				if (clntinfo->havetoconvert != CONV_NOTCONVERT) {
+					char* tmp;
+					if (clntinfo->havetoconvert
+							== CONV_TOASCII) {
+						tmp = to_ascii(buffer, &count);
+						if (count > TRANSMITBUFSIZE) {
+							buffer =
+							realloc(buffer, count);
+						}
+						memcpy((void*) buffer,
+							(void*) tmp, count);
+						free(tmp);
+					}
+					if (clntinfo->havetoconvert
+							== CONV_FRMASCII) {
+						tmp = from_ascii(buffer,
+								&count);
+						buffer = tmp;
+					}
+				}
 				nwritten = write(clntinfo->dataclientsock,
 						pbuf, count);
 				if (nwritten < 0) {
-					log(8, "Error writing (%s, %d): %s",
+					log(2, "Error writing (%s, %d): %s",
 							__FILE__, __LINE__,
 							strerror(errno));
 					break;
@@ -833,10 +723,21 @@ int transmitfile(struct clientinfo *clntinfo)  {
 					pbuf = 0;
 				}
 			} while (pbuf);
+			/* break's from the above while - loop go here, pass
+			 * them along */
+			if (nwritten < 0) {
+				break;
+			}
+			if (scret <= 0) {
+				break;
+			}
 		}
 		if (FD_ISSET(cs, &exceptset) || FD_ISSET(cs, &readset)) {
 			if (checkforabort(clntinfo)) {
-				error = 2;
+				error = TRNSMT_NOERRORMSG;
+				break;
+			} else {
+				aborted = 1;
 				break;
 			}
 		}
@@ -849,51 +750,60 @@ int transmitfile(struct clientinfo *clntinfo)  {
 		FD_SET(cs, &exceptset);
 		FD_SET(cs, &readset);
 		/* Linux modifies the timeout struct */
-		readtime.tv_sec = 300;
+		readtime.tv_sec = config_get_ioption("transfertimeout", 300);
 		readtime.tv_usec = 0;
-		writetime.tv_sec = 300;
+		writetime.tv_sec = config_get_ioption("transfertimeout", 300);
 		writetime.tv_usec = 0;
 	}
 
 	/* Restore flags */
-	ret = fcntl(clntinfo->dataclientsock, F_SETFL, fdflags);
-	if (ret < 0) {
-		log(2, "Error resetting fd flags");
+	if (!aborted) {
+		ret = fcntl(clntinfo->dataclientsock, F_SETFL, fdflags);
+		if (ret < 0) {
+			log(2, "Error resetting fd flags");
+		}
 	}
 
 	if (sret == -1 || scret == -1) {
 		log(2, "Error in select() occured: %s", strerror(errno));
-		li.respcode = 500;
+		lcs.respcode = 500;
 		say(cs, "500 Error in select() occured\r\n");
-		error = 2;  /* do not generate an error message */
+		error = TRNSMT_NOERRORMSG;
 	}
 	/* count == 0 means correct end of transfer */
 	if ((sret == 0 || scret == 0) && count != 0) {
-		log(2, "Connection timed out in transmitfile(): %d, %d",
+		log(2, "Connection timed out in transfer_transmit(): %d, %d",
 				readtime.tv_sec, writetime.tv_sec);
 		log(9, "Sockets: dataserver: %d, dataclient: %d, client: %d",
 				clntinfo->dataserversock,
 				clntinfo->dataclientsock,
-				cs);
+				clntinfo->clientsocket);
 		if (scret == 0) {
 			log(9, "scret == 0 (write)");
 		}
 		if (sret == 0) {
 			log(9, "sret == 0 (read)");
 		}
-		error = 1;
+		error = TRNSMT_ERROR;
 	}
 
+	lcs.transferred = totwritten;
+	log(7, "Transferred %d bytes", totwritten);
+
+	free(buffer);
+
 	close(clntinfo->dataclientsock);
 	close(clntinfo->dataserversock);
+	if (clntinfo->cachefd >= 0) {
+		close(clntinfo->cachefd);
+	}
 	clntinfo->dataclientsock = -1;
 	clntinfo->dataserversock = -1;
+	clntinfo->cachefd        = -1;
 
-	li.transferred = totwritten;
-	log(7, "Transferred %d bytes", totwritten);
-	
-	free(buffer);
-	clntinfo->havetoconvert = CONV_NOTCONVERT;
+	if (aborted) {
+		error = TRNSMT_ABORTED;
+	}
 	return error;
 }
 
@@ -907,9 +817,12 @@ int checkforabort(struct clientinfo* clntinfo) {
 	char dmseq[1]  = { (char) DM };
 	int ret;
 
-	struct timeval tm = { 300, 0 };
+	struct timeval tm;
 	fd_set set, exset;
 
+	tm.tv_sec = config_get_ioption("comamndtimeout", 300);
+	tm.tv_usec = 0;
+
 	/* this read won't block */
 	ret = read(clientfd, buffer, sizeof(buffer) - 1);
 	if (ret < 0) {
@@ -951,7 +864,7 @@ int checkforabort(struct clientinfo* clntinfo) {
 	buffer[ret] = '\0';
 
 	if(buffer[0] != (char) DM ||
-		strncasecmp(buffer + 1,"ABOR",4) != 0) {
+		strncasecmp(buffer + 1, "ABOR", 4) != 0) {
 		return 0;
 	}
 
@@ -975,21 +888,12 @@ int checkforabort(struct clientinfo* clntinfo) {
 		 */
 		return 1;
 	}
-#ifdef SFTP_SUPPORT
-	if (clntinfo->servermode == SECURE_SFTP) {
-		/* just close the connection */
-		cancelled = 1;
-		do_cancel();
-		close(clntinfo->dataserversock);
-		return 1;
-	}
-#endif
 
 	if (send(serverfd, sabseq, 3, MSG_OOB) != 3
-		|| send(serverfd, dmseq, 1, 0) != 1) {
+		|| send(serverfd, dmseq, 1, 0) != 1
+		|| send(serverfd, "ABOR\r\n", 6, 0) != 6) {
 		log(3, "send() error: %s", strerror(errno));
 	}
-	say(serverfd, "ABOR\r\n");
 
 
 	FD_ZERO(&set);
@@ -1007,16 +911,19 @@ int checkforabort(struct clientinfo* clntinfo) {
 	ret = ftp_getrc(serverfd, &buf);
 	switch (ret) {
 		case 425:
+		case 451:
 		case 426:
-			ret = ftp_getrc(serverfd, 0);
-			li.respcode = ret;
+			free(buf);
+			ret = ftp_getrc(serverfd, &buf);
+			lcs.respcode = ret;
 			/* fallthrough */
 		case 226:
 			log(6, "Client requested ABORT - Aborted successfully");
-			break;
+			free(buf);
+			return 0;
 		default:
 			log(6, "Answer of server after ABORT unknown: (%s).",
-					buffer);
+					buf);
 	}
 	free(buf);
 	return 1;
