@@ -36,15 +36,32 @@
 #ifndef _LOG_H
 #define _LOG_H
 
-extern void log(int, char *fmt, ...);
+extern void log(int, const char *fmt, ...);
 
-struct loginfo {
-	char* client;
-	char* server;
-	char* user;
-	char* cmd;
-	char* name;
-	char* service;   /* either "ftp" or "sftp"          */
+/* used:
+ *          a A b C d (e) f F h H i I l L m n p P r R s t T u v        */
+
+struct log_cmd_st {
+
+	const char* svrname;
+	char* svrip;
+	const char* svrlogin;
+
+	const char* clntname;
+	char* clntip;
+
+	char* ifipclnt;
+	char* ifipsvr;
+
+	char* userlogin;
+	char* usereffective;
+	char* userforwarded;
+	const char* anon_user;
+
+	const char* cmd;
+	char* method;
+	const char* filename;
+	const char* service;   /* "ftp" */
 	char direction;  /* either 'o'utgoing or 'i'ncoming */
 	char type;	 /* either 'i'mage or 'a'scii       */
 	int respcode;
@@ -53,6 +70,28 @@ struct loginfo {
 	unsigned long int transferred;
 };
 
-extern void log_cmd(struct loginfo*);
+
+struct loginfo_st {
+	int syslog;
+	char* syslog_facility;
+	int debuglevel;
+	char* logf_name;
+	FILE* logf;
+	struct cmdlogent_t {
+		/* list of opened logfiles */
+		char* logf_name;
+		int logf_size;  /* not used for files but for dirs */
+		char* specs;
+		FILE* logf;
+		char* style;
+		struct cmdlogent_t* next;
+	} *cmdlogfiles, *cmdlogdirs;
+};
+
+
+void log_cmd(struct log_cmd_st*);
+int log_init(void);
+int log_detect_log_change(void);
+
 
 #endif
