@@ -16,26 +16,32 @@
  * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
  */
 
-int std_quit(char*, struct conn_info_st*);
-int std_user(char*, struct conn_info_st*);
-int std_pass(char*, struct conn_info_st*);
-int std_pasv(char*, struct conn_info_st*);
-int std_port(char*, struct conn_info_st*);
-int std_stor(char*, struct conn_info_st*);
-int std_retr(char*, struct conn_info_st*);
-int std_type(char*, struct conn_info_st*);
+int std_loggedin(const char*, struct conn_info_st*);
+int std_quit(const char*, struct conn_info_st*);
+int std_pasv(const char*, struct conn_info_st*);
+int std_epsv(const char*, struct conn_info_st*);
+int std_port(const char*, struct conn_info_st*);
+int std_stor(const char*, struct conn_info_st*);
+int std_retr(const char*, struct conn_info_st*);
+int std_type(const char*, struct conn_info_st*);
+int std_list(const char*, struct conn_info_st*);
 
 
 struct cmdhandlerstruct std_cmdhandler[] = {
-	{ "USER ", std_user },
-	{ "PASS ", std_pass },
+	{ "USER ", std_loggedin },
+	{ "PASS ", std_loggedin },
 	{ "PORT ", std_port },
 	{ "PASV",  std_pasv },
+	{ "EPSV",  std_epsv },
 	{ "STOR ", std_stor },
 	{ "RETR ", std_retr },
+	{ "APPE ", std_stor },
 	{ "TYPE ", std_type },
 	{ "QUIT", std_quit },
+	{ "LIST", std_list },
+	{ "NLST", std_list },
 	{ 0, 0 }
 };
 
 
+
